from crispy_forms.helper import FormHelper
from crispy_forms.layout import HTML, Div, Fieldset, Layout, Submit
from django import forms
from django.utils.translation import gettext_lazy as _


class AdminCrispyForm(forms.Form):
    """
    Base crispy form to be used in custom views within the django admin.
    """

    section_title = _('No title defined')
    button_label = _('Save')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Build fieldset
        fieldset_list = ['']
        for field in self.fields:
            fieldset_list.append(Div(field, css_class='form-row field-name'))

        # Crispy
        self.helper = FormHelper()
        self.helper.form_method = 'post'
        self.helper.add_input(Submit('submit', self.button_label, css_class="button btn-primary"))
        self.helper.layout = Layout(
            Div(
                Div(HTML(f'<h2>{self.section_title}</h2>'), Fieldset(*fieldset_list), css_class='module aligned'),
                css_class='custom-form',
            ),
        )



from django.test import TestCase

from ai_django_core.utils import object_to_dict
from testapp.models import MySingleSignalModel


class UtilModelTest(TestCase):
    def test_object_to_dict_regular(self):
        obj = MySingleSignalModel.objects.create(value=17)
        self.assertEqual(object_to_dict(obj), {'value': obj.value})

    def test_object_to_dict_blacklist(self):
        obj = MySingleSignalModel.objects.create(value=17)
        self.assertEqual(object_to_dict(obj, ['value']), {})

    def test_object_to_dict_with_id_with_blacklist(self):
        obj = MySingleSignalModel.objects.create(value=17)
        self.assertEqual(object_to_dict(obj, ['value'], True), {'id': obj.id})

    def test_object_to_dict_with_id_no_blacklist(self):
        obj = MySingleSignalModel.objects.create(value=17)
        self.assertEqual(object_to_dict(obj, include_id=True), {'id': obj.id, 'value': obj.value})



from unittest import mock

from django.test import TestCase

from testapp.models import ModelWithCleanMixin


class CleanOnSaveMixinTest(TestCase):
    def test_clean_is_called(self):
        obj = ModelWithCleanMixin()
        with mock.patch.object(obj, 'clean') as mocked_method:
            obj.save()

        mocked_method.assert_called_once()



from os.path import basename

from django.conf import settings
from django.core.mail import EmailMultiAlternatives
from django.test import TestCase, override_settings
from freezegun import freeze_time

from ai_django_core.mail.errors import EmailServiceAttachmentError, EmailServiceConfigError
from ai_django_core.mail.services.base import BaseEmailService, BaseEmailServiceFactory


class BaseEmailServiceFactoryTest(TestCase):
    class TestMailService(BaseEmailService):
        subject = 'My subject'
        template_name = 'test_email.html'

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

    def test_init_recipient_list_is_set(self):
        email = 'albertus.magnus@example.com'
        factory = BaseEmailServiceFactory([email])
        self.assertEqual(factory.recipient_email_list, [email])

    def test_is_valid_positive_case(self):
        email = 'albertus.magnus@example.com'
        factory = BaseEmailServiceFactory(recipient_email_list=[email])
        factory.service_class = BaseEmailService
        self.assertTrue(factory.is_valid())

    def test_is_valid_no_recipients(self):
        factory = BaseEmailServiceFactory()
        factory.service_class = BaseEmailService
        with self.assertRaises(EmailServiceConfigError):
            factory.is_valid()

    def test_is_valid_no_service_class(self):
        email = 'albertus.magnus@example.com'
        factory = BaseEmailServiceFactory(recipient_email_list=[email])
        with self.assertRaises(EmailServiceConfigError):
            factory.is_valid()

    def test_is_valid_no_exception_raised(self):
        factory = BaseEmailServiceFactory()
        factory.is_valid(raise_exception=False)
        self.assertEqual(len(factory.errors), 2)

    def test_has_errors_positive_case(self):
        factory = BaseEmailServiceFactory()
        factory.is_valid(raise_exception=False)
        self.assertTrue(factory.has_errors())

    def test_has_errors_negative_case(self):
        email = 'albertus.magnus@example.com'
        factory = BaseEmailServiceFactory(recipient_email_list=[email])
        factory.service_class = BaseEmailService
        self.assertFalse(factory.has_errors())

    def test_get_recipient_list_regular(self):
        email_1 = 'albertus.magnus@example.com'
        email_2 = 'thomas.von.aquin@example.com'
        factory = BaseEmailServiceFactory(recipient_email_list=[email_1, email_2])
        self.assertEqual(factory.get_recipient_list(), [email_1, email_2])

    def test_get_email_from_recipient_regular(self):
        email_1 = 'albertus.magnus@example.com'
        email_2 = 'thomas.von.aquin@example.com'
        factory = BaseEmailServiceFactory(recipient_email_list=[email_1, email_2])
        self.assertEqual(factory.get_email_from_recipient(factory.get_recipient_list()[1]), email_2)

    def test_get_context_data_regular(self):
        factory = BaseEmailServiceFactory()
        self.assertEqual(factory.get_context_data(), {})

    def test_process_regular(self):
        email_1 = 'albertus.magnus@example.com'
        email_2 = 'thomas.von.aquin@example.com'
        factory = BaseEmailServiceFactory(recipient_email_list=[email_1, email_2])
        factory.service_class = self.TestMailService
        self.assertEqual(factory.process(), 2)

    def test_process_with_excepetion(self):
        factory = BaseEmailServiceFactory()
        factory.service_class = self.TestMailService
        with self.assertRaises(EmailServiceConfigError):
            factory.process()


class BaseEmailServiceTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

    def test_init_recipient_as_str_is_wrapped_to_list(self):
        email = 'albertus.magnus@example.com'
        service = BaseEmailService(email)
        self.assertEqual(service.recipient_email_list, [email])

    def test_init_recipient_and_context_are_initialised_empty(self):
        service = BaseEmailService()
        self.assertEqual(service.recipient_email_list, [])
        self.assertEqual(service.context_data, {})

    def test_get_context_data_regular(self):
        data = {'city': 'Cologne'}
        service = BaseEmailService(context_data=data)
        self.assertEqual(service.get_context_data(), data)

    def test_get_subject_no_prefix(self):
        subject = 'I am a subject!'
        service = BaseEmailService()
        service.subject = subject
        self.assertEqual(service.get_subject(), subject)

    def test_get_subject_with_prefix(self):
        prefix = 'Ai: Core'
        subject = 'I am a subject!'
        service = BaseEmailService()
        service.SUBJECT_PREFIX = prefix
        service.subject = subject
        self.assertIn(prefix, service.get_subject())
        self.assertIn(subject, service.get_subject())

    @override_settings(DEFAULT_FROM_EMAIL='noreply@example.com')
    def test_get_from_email_from_settings(self):
        service = BaseEmailService()
        self.assertEqual(service.get_from_email(), 'noreply@example.com')

    @override_settings(DEFAULT_FROM_EMAIL='noreply@example.com')
    def test_get_from_email_from_class(self):
        email = 'albertus.magnus@example.com'
        service = BaseEmailService()
        service.FROM_EMAIL = email
        self.assertEqual(service.get_from_email(), email)

    def test_get_reply_to_email_not_set(self):
        service = BaseEmailService()
        self.assertEqual(service.get_reply_to_emails(), [])

    def test_get_reply_to_email_is_set(self):
        email = 'albertus.magnus@example.com'
        service = BaseEmailService()
        service.REPLY_TO_ADDRESS = email
        self.assertEqual(service.get_reply_to_emails(), [email])

    def test_get_cc_email_not_set(self):
        service = BaseEmailService()
        self.assertEqual(service.get_cc_emails(), [])

    def test_get_cc_email_is_set(self):
        email = 'albertus.magnus@example.com'
        service = BaseEmailService()
        service.cc_email_list = [email]
        self.assertEqual(service.get_cc_emails(), [email])

    def test_get_bcc_email_not_set(self):
        service = BaseEmailService()
        self.assertEqual(service.get_bcc_emails(), [])

    def test_get_bcc_email_is_set(self):
        email = 'albertus.magnus@example.com'
        service = BaseEmailService()
        service.bcc_email_list = [email]
        self.assertEqual(service.get_bcc_emails(), [email])

    @override_settings(LANGUAGE_CODE='de-AT')
    def test_get_translation_regular_german(self):
        service = BaseEmailService()
        self.assertEqual(service.get_translation(), 'de')

    @override_settings(LANGUAGE_CODE='en-GB')
    def test_get_translation_regular_english(self):
        service = BaseEmailService()
        self.assertEqual(service.get_translation(), 'en')

    @override_settings(LANGUAGE_CODE='de')
    def test_get_translation_settings_short(self):
        service = BaseEmailService()
        self.assertEqual(service.get_translation(), 'de')

    @override_settings(LANGUAGE_CODE=None)
    def test_get_translation_settings_not_set(self):
        service = BaseEmailService()
        self.assertEqual(service.get_translation(), None)

    @override_settings(LANGUAGE_CODE=1)
    def test_get_translation_settings_invalid_type(self):
        service = BaseEmailService()
        self.assertEqual(service.get_translation(), None)

    @override_settings(LANGUAGE_CODE='a')
    def test_get_translation_settings_invalid_value(self):
        service = BaseEmailService()
        self.assertEqual(service.get_translation(), None)

    def test_get_attachments_regular(self):
        file_path = 'usr/albertus/myfile.csv'
        service = BaseEmailService(attachment_list=[file_path])
        self.assertEqual(service.get_attachments(), [file_path])

    def test_get_attachments_empty(self):
        service = BaseEmailService()
        self.assertEqual(service.get_attachments(), [])

    def test_add_attachments_path_as_str(self):
        # Setup
        service = BaseEmailService()
        service.template_name = 'test_email.html'
        msg_obj = service._build_mail_object()

        file_path = settings.BASE_PATH / 'tests/files/testfile.txt'
        service.attachment_list = [file_path]
        msg_obj = service._add_attachments(msg_obj)

        self.assertEqual(len(msg_obj.attachments), 1)
        self.assertEqual(msg_obj.attachments[0][0], basename(file_path))

    def test_add_attachments_path_as_dict(self):
        # Setup
        service = BaseEmailService()
        service.template_name = 'test_email.html'
        msg_obj = service._build_mail_object()

        filename = 'awesome_file.txt'
        file_path = settings.BASE_PATH / 'tests/files/testfile.txt'
        service.attachment_list = [{'filename': filename, 'file': file_path, 'mimetype': 'text/text'}]
        msg_obj = service._add_attachments(msg_obj)

        self.assertEqual(len(msg_obj.attachments), 1)
        self.assertEqual(msg_obj.attachments[0][0], filename)

    def test_add_attachments_path_wrong_dict_data(self):
        # Setup
        service = BaseEmailService()
        service.template_name = 'test_email.html'
        msg_obj = service._build_mail_object()

        filename = 'awesome_file.txt'
        service.attachment_list = [{'filename': filename}]

        with self.assertRaises(EmailServiceAttachmentError):
            service._add_attachments(msg_obj)

    @freeze_time('2020-06-26')
    def test_build_mail_object_regular(self):
        from_email = 'noreply@example.com'
        reply_to_email = 'willreply@example.com'
        email = 'albertus.magnus@example.com'
        subject = 'Test email'
        my_var = 'Lorem ipsum dolor!'
        service = BaseEmailService(recipient_email_list=[email], context_data={'my_var': my_var})
        service.FROM_EMAIL = from_email
        service.REPLY_TO_ADDRESS = reply_to_email
        service.subject = subject
        service.template_name = 'test_email.html'
        msg_obj = service._build_mail_object()

        # Assertions
        self.assertIsInstance(msg_obj, EmailMultiAlternatives)

        self.assertEqual(msg_obj.subject, subject)

        self.assertEqual(msg_obj.from_email, from_email)
        self.assertEqual(msg_obj.to, [email])
        self.assertEqual(msg_obj.reply_to, [reply_to_email])

        self.assertIn('Friday', msg_obj.body)
        self.assertIn(my_var, msg_obj.body)

        self.assertIn('Friday', msg_obj.alternatives[0][0])
        self.assertIn(my_var, msg_obj.alternatives[0][0])

    def test_build_mail_object_with_attachments(self):
        from_email = 'noreply@example.com'
        reply_to_email = 'willreply@example.com'
        email = 'albertus.magnus@example.com'
        subject = 'Test email'
        my_var = 'Lorem ipsum dolor!'
        file_path = settings.BASE_PATH / 'tests/files/testfile.txt'
        service = BaseEmailService(
            recipient_email_list=[email], context_data={'my_var': my_var}, attachment_list=[file_path]
        )
        service.FROM_EMAIL = from_email
        service.REPLY_TO_ADDRESS = reply_to_email
        service.subject = subject
        service.template_name = 'test_email.html'
        msg_obj = service._build_mail_object()

        # Assertions
        self.assertIsInstance(msg_obj, EmailMultiAlternatives)

        self.assertEqual(len(msg_obj.attachments), 1)
        self.assertEqual(msg_obj.attachments[0][0], basename(file_path))

    def test_setting_txt_templates_works(self):
        my_var = 'Lorem ipsum dolor!'
        service = BaseEmailService(
            recipient_email_list=['albertus.magnus@example.com'], context_data={'my_var': my_var}
        )
        service.FROM_EMAIL = 'noreply@example.com'
        service.subject = 'Test mail'
        service.template_name = 'test_email.html'
        service.template_txt_name = 'test_email.txt'
        msg_obj = service._build_mail_object()

        # Assertions
        self.assertIsInstance(msg_obj, EmailMultiAlternatives)

        self.assertIn('I am a different content', msg_obj.body)
        self.assertNotIn('I am a different content', msg_obj.alternatives[0][0])

        self.assertIn(my_var, msg_obj.body)

    @freeze_time('2020-06-26')
    @override_settings(LANGUAGE_CODE='nl-BE')
    def test_build_mail_object_translation_works(self):
        service = BaseEmailService(recipient_email_list='noreply@example.com')
        service.template_name = 'test_email.html'
        msg_obj = service._build_mail_object()

        # Assertions
        self.assertIsInstance(msg_obj, EmailMultiAlternatives)

        self.assertIn('vrijdag', msg_obj.body)
        self.assertIn('vrijdag', msg_obj.alternatives[0][0])

    def test_is_valid_positive_case(self):
        email = 'albertus.magnus@example.com'
        subject = 'Test email'
        service = BaseEmailService(recipient_email_list=[email])
        service.subject = subject
        service.template_name = 'test_email.html'
        self.assertTrue(service.is_valid())

    def test_is_valid_no_subject(self):
        email = 'albertus.magnus@example.com'
        service = BaseEmailService(recipient_email_list=[email])
        service.template_name = 'test_email.html'
        with self.assertRaises(EmailServiceConfigError):
            service.is_valid()

    def test_is_valid_no_template(self):
        email = 'albertus.magnus@example.com'
        subject = 'Test email'
        service = BaseEmailService(recipient_email_list=[email])
        service.subject = subject
        with self.assertRaises(EmailServiceConfigError):
            service.is_valid()

    def test_is_valid_no_recipient(self):
        subject = 'Test email'
        service = BaseEmailService()
        service.subject = subject
        with self.assertRaises(EmailServiceConfigError):
            service.is_valid()

    def test_is_valid_no_exception_raised(self):
        service = BaseEmailService()
        service.is_valid(raise_exception=False)
        self.assertEqual(len(service.errors), 3)

    def test_has_errors_positive_case(self):
        service = BaseEmailService()
        service.is_valid(raise_exception=False)
        self.assertTrue(service.has_errors())

    def test_has_errors_negative_case(self):
        email = 'albertus.magnus@example.com'
        subject = 'Test email'
        service = BaseEmailService(recipient_email_list=[email])
        service.subject = subject
        service.template_name = 'test_email.html'
        self.assertFalse(service.has_errors())

    def test_process_regular(self):
        email = 'albertus.magnus@example.com'
        subject = 'Test email'
        service = BaseEmailService(recipient_email_list=[email])
        service.subject = subject
        service.template_name = 'test_email.html'
        self.assertTrue(service.process())

    def test_process_with_error(self):
        subject = 'Test email'
        service = BaseEmailService()
        service.subject = subject
        service.template_name = 'test_email.html'
        with self.assertRaises(EmailServiceConfigError):
            service.process()



from unittest import mock

from django.contrib import admin
from django.contrib.auth.models import User
from django.http import HttpResponseRedirect
from django.test import TestCase

from ai_django_core.admin.model_admins.mixins import DeactivatableChangeViewAdminMixin
from ai_django_core.tests.mixins import RequestProviderMixin


class TestAdmin(DeactivatableChangeViewAdminMixin, admin.ModelAdmin):
    pass


class DeactivatableChangeViewAdminMixinTest(RequestProviderMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        # Use random model for this meta test
        cls.admin = TestAdmin(admin_site=None, model=User)
        cls.user = User.objects.create(username="test_user", is_superuser=False)
        cls.super_user = User.objects.create(username="super_user", is_superuser=True)

    def setUp(self) -> None:
        super().setUp()
        self.admin.enable_change_view = True

    def test_can_see_change_view_positive_flag(self):
        self.assertTrue(self.admin.can_see_change_view(request=self.get_request()))

    def test_can_see_change_view_negative_flag(self):
        self.admin.enable_change_view = False
        self.assertFalse(self.admin.can_see_change_view(request=self.get_request()))

    def test_get_list_display_links_can_see_method_called(self):
        with mock.patch.object(self.admin, 'can_see_change_view', return_value=True) as mock_method:
            self.admin.get_list_display_links(request=self.get_request(user=self.user), list_display=('first_name',))

        mock_method.assert_called_once()

    def test_get_list_display_links_can_see_method_positive_flag(self):
        field_tuple = ('first_name',)
        self.assertEqual(
            list(field_tuple),
            self.admin.get_list_display_links(request=self.get_request(user=self.user), list_display=field_tuple),
        )

    def test_get_list_display_links_can_see_method_negative_flag(self):
        self.admin.enable_change_view = False
        self.assertIsNone(
            self.admin.get_list_display_links(request=self.get_request(user=self.user), list_display=('first_name',))
        )

    def test_change_view_can_see_method_called_because_of_positive_flag(self):
        with mock.patch.object(self.admin, 'can_see_change_view', return_value=True) as mocked_can_see_method:
            with mock.patch('django.contrib.admin.ModelAdmin.change_view') as mocked_base_change_view:
                self.admin.change_view(request=self.get_request(user=self.super_user), object_id=str(self.user.id))

        mocked_can_see_method.assert_called_once()
        mocked_base_change_view.assert_called_once()

    def test_change_view_can_see_method_not_called_because_of_negative_flag(self):
        with mock.patch.object(self.admin, 'can_see_change_view', return_value=False) as mocked_can_see_method:
            with mock.patch('django.contrib.admin.ModelAdmin.change_view') as mocked_base_change_view:
                self.admin.change_view(request=self.get_request(user=self.super_user), object_id=str(self.user.id))

        mocked_can_see_method.assert_called_once()
        mocked_base_change_view.assert_not_called()

    def test_change_view_can_see_method_not_called_but_redirect(self):
        self.admin.enable_change_view = False
        result = self.admin.change_view(request=self.get_request(user=self.super_user), object_id=str(self.user.id))

        self.assertIsInstance(result, HttpResponseRedirect)



from django.conf import settings
from django.db import models
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from ai_django_core.middleware.current_user import CurrentUserMiddleware


class CreatedAtInfo(models.Model):
    created_at = models.DateTimeField(_("Created at"), default=now, db_index=True)

    def save(self, *args, **kwargs):
        # just a fallback for old data
        if not self.created_at:
            self.created_at = now()
        super().save(*args, **kwargs)

    class Meta:
        abstract = True


class CommonInfo(CreatedAtInfo, models.Model):
    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("Created by"),
        blank=True,
        null=True,
        related_name="%(app_label)s_%(class)s_created",
        on_delete=models.SET_NULL,
    )
    lastmodified_at = models.DateTimeField(_("Last modified at"), default=now, db_index=True)
    lastmodified_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("Last modified by"),
        blank=True,
        null=True,
        related_name="%(app_label)s_%(class)s_lastmodified",
        on_delete=models.SET_NULL,
    )

    @staticmethod
    def get_current_user():
        """
        Get the currently logged in user over middleware.
        Can be overwritten to use e.g. other middleware or additional functionality.
        :return: user instance
        """
        return CurrentUserMiddleware.get_current_user()

    def set_user_fields(self, user):
        """
        Set user-related fields before saving the instance.
        If no user with primary key is given the fields are not set.
        :param user: user instance of current user
        """
        if user and user.pk:
            if not self.pk:
                self.created_by = user
            self.lastmodified_by = user

    def save(self, *args, **kwargs):
        self.lastmodified_at = now()
        current_user = self.get_current_user()
        self.set_user_fields(current_user)
        super().save(*args, **kwargs)

    class Meta:
        abstract = True



from django.test import TestCase, override_settings

from ai_django_core.services.custom_scrubber import AbstractScrubbingService


class AbstractScrubbingServiceTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.service = AbstractScrubbingService()

    @override_settings(DEBUG=False)
    def test_scrubber_debug_mode_needs_to_be_active(self):
        self.assertEqual(self.service.process(), False)

    @override_settings(DEBUG=True, INSTALLED_APPS=[])
    def test_scrubber_needs_to_be_installed(self):
        self.assertEqual(self.service.process(), False)

    # todo write more tests



import os

from django import template
from django.conf import settings

register = template.Library()


@register.filter
def filename(value, max_length=25):
    """
    Shortens the filename to maxlength 25 without loosing the file extension

    :param value:
    :param max_length:
    :return filename with a max length of 25
    """
    name = os.path.basename(value.url)
    if len(name) > max_length:
        ext = name.split('.')[-1]
        name = "%s[..].%s" % (name[:max_length], ext)
    return name


@register.filter
def filesize(value):
    """
    Returns the filesize of the filename given in value

    :param value:
    :return filesize:
    """
    try:
        return os.path.getsize("%s%s" % (settings.MEDIA_ROOT, value))
    except Exception:
        return 0



import datetime

import pytz
from django.test import TestCase

from ai_django_core.utils.string import (
    date_to_string,
    datetime_to_string,
    distinct,
    encode_to_xml,
    float_to_string,
    number_to_string,
    slugify_file_name,
    smart_truncate,
    string_or_none_to_string,
)


class UtilsStringTest(TestCase):
    def test_distinct_regular(self):
        not_distinct_list = ['Beer', 'Wine', 'Whiskey', 'Beer']
        distinct_list = distinct(not_distinct_list)

        self.assertEqual(len(distinct_list), 3)
        self.assertIn('Beer', distinct_list)
        self.assertIn('Whiskey', distinct_list)
        self.assertIn('Wine', distinct_list)

    def test_slugify_file_name_regular(self):
        filename = 'hola and hello.txt'
        slug = slugify_file_name(filename)
        self.assertEqual(slug, 'hola_and_hello.txt')

    def test_slugify_file_name_nothing_to_slugify(self):
        filename = 'hola.txt'
        slug = slugify_file_name(filename)
        self.assertEqual(slug, filename)

    def test_slugify_file_name_max_length(self):
        filename = 'a very long filename.txt'
        slug = slugify_file_name(filename, 6)
        self.assertEqual(slug, 'a_very.txt')

    def test_smart_truncate_in_word(self):
        my_sentence = 'I am a very interesting sentence.'
        truncated_str = smart_truncate(my_sentence, 10)
        self.assertEqual(truncated_str, 'I am a...')

    def test_smart_truncate_after_word(self):
        my_sentence = 'I am a very interesting sentence.'
        truncated_str = smart_truncate(my_sentence, 14)
        self.assertEqual(truncated_str, 'I am a very...')

    def test_smart_truncate_changed_postfix(self):
        my_sentence = 'I am a very interesting sentence.'
        truncated_str = smart_truncate(my_sentence, 10, '[...]')
        self.assertEqual(truncated_str, 'I am a[...]')

    def test_smart_truncate_not_cutting_on_too_short_strings(self):
        my_sentence = 'I am a very interesting sentence.'
        truncated_str = smart_truncate(my_sentence, 100, '---')
        self.assertEqual(truncated_str, my_sentence)

    def test_float_to_string_regular(self):
        self.assertEqual(float_to_string(5.61), '5,61')

    def test_float_to_string_value_replacement_not_used(self):
        self.assertEqual(float_to_string(4.41, '-'), '4,41')

    def test_float_to_string_no_value_replacement_used(self):
        self.assertEqual(float_to_string(None, 'Heureka'), 'Heureka')

    def test_float_to_string_value_greater_thousand(self):
        self.assertEqual(float_to_string(1234.56), '1234,56')

    def test_date_to_string_regular(self):
        self.assertEqual(date_to_string(datetime.date(2020, 9, 19)), '19.09.2020')

    def test_date_to_string_other_format(self):
        self.assertEqual(date_to_string(datetime.date(2020, 9, 19), str_format='%Y-%m-%d'), '2020-09-19')

    def test_date_to_string_replacement_undefined(self):
        self.assertEqual(date_to_string(None), '-')

    def test_date_to_string_replacement_defined(self):
        self.assertEqual(date_to_string(None, 'no date'), 'no date')

    def test_datetime_to_string_regular(self):
        self.assertEqual(datetime_to_string(datetime.datetime(2020, 9, 19, 8, tzinfo=pytz.UTC)), '19.09.2020 08:00')

    def test_datetime_to_string_other_format(self):
        self.assertEqual(
            datetime_to_string(datetime.datetime(2020, 9, 19, 8, tzinfo=pytz.UTC), str_format='%Y-%m-%d'), '2020-09-19'
        )

    def test_datetime_to_string_replacement_undefined(self):
        self.assertEqual(datetime_to_string(None), '-')

    def test_datetime_to_string_replacement_defined(self):
        self.assertEqual(datetime_to_string(None, 'no date'), 'no date')

    def test_number_to_string_regular(self):
        self.assertEqual(number_to_string(5.61, decimal_digits=2), '5.61')

    def test_number_to_string_value_replacement_not_used(self):
        self.assertEqual(number_to_string(4.41, decimal_digits=2, replacement='-'), '4.41')

    def test_number_to_string_no_value_replacement_used(self):
        self.assertEqual(number_to_string(None, replacement='Heureka'), 'Heureka')

    def test_number_to_string_value_greater_thousand(self):
        self.assertEqual(number_to_string(1234.56, decimal_digits=2), '1,234.56')

    def test_number_to_string_int_value_no_digits(self):
        self.assertEqual(number_to_string(117), '117')

    def test_number_to_string_int_value_with_digits(self):
        self.assertEqual(number_to_string(117, decimal_digits=2), '117.00')

    def test_string_or_none_to_string_regular(self):
        my_str = 'I am a string.'
        self.assertEqual(string_or_none_to_string(my_str), my_str)

    def test_string_or_none_to_string_replacement_undefined(self):
        self.assertEqual(string_or_none_to_string(None), '-')

    def test_string_or_none_to_string_replacement_defined(self):
        self.assertEqual(string_or_none_to_string(None, 'no value'), 'no value')

    def test_encode_to_xml_regular(self):
        xml_str = '<tag>Something with an ampersand (&)</tag>'
        self.assertEqual(encode_to_xml(xml_str), '&lt;tag&gt;Something with an ampersand (&amp;)&lt;/tag&gt;')



from django.test import TestCase
from rest_framework import serializers
from rest_framework.serializers import ListSerializer

from ai_django_core.drf.fields import RecursiveField
from testapp.models import ModelWithFkToSelf, ModelWithOneToOneToSelf


class TestManyTrueSerializer(serializers.ModelSerializer):
    children = RecursiveField(many=True)

    class Meta:
        model = ModelWithFkToSelf
        fields = [
            'id',
            'children',
        ]


class TestManyFalseSerializer(serializers.ModelSerializer):
    peer = RecursiveField()

    class Meta:
        model = ModelWithOneToOneToSelf
        fields = [
            'id',
            'peer',
        ]


class RecursiveFieldTest(TestCase):
    def test_many_true_regular(self):
        serializer = TestManyTrueSerializer()

        self.assertIn('children', serializer.fields)
        self.assertIsInstance(serializer.fields['children'], ListSerializer)
        self.assertIsInstance(serializer.fields['children'].child, RecursiveField)

    def test_many_true_representation(self):
        mwfts_1 = ModelWithFkToSelf.objects.create(parent=None)
        mwfts_2 = ModelWithFkToSelf.objects.create(parent=mwfts_1)

        serializer = TestManyTrueSerializer(instance=mwfts_1)
        representation = serializer.to_representation(instance=mwfts_1)

        self.assertIsInstance(representation, dict)
        self.assertIn('children', representation)
        self.assertEqual(len(representation['children']), 1)
        self.assertEqual(representation['children'][0]['id'], mwfts_2.id)
        self.assertEqual(representation['children'][0]['children'], [])

    def test_many_false_regular(self):
        serializer = TestManyFalseSerializer()

        self.assertIn('peer', serializer.fields)
        self.assertIsInstance(serializer.fields['peer'], RecursiveField)

    def test_many_false_representation(self):
        mwotos_no_peer = ModelWithOneToOneToSelf.objects.create(peer=None)
        mwotos_has_peer = ModelWithOneToOneToSelf.objects.create(peer=mwotos_no_peer)

        serializer = TestManyFalseSerializer(instance=mwotos_has_peer)
        representation = serializer.to_representation(instance=mwotos_has_peer)

        self.assertIsInstance(representation, dict)
        self.assertIn('peer', representation)
        self.assertEqual(len(representation['peer']), 2)
        self.assertEqual(representation['peer']['id'], mwotos_no_peer.id)
        self.assertEqual(representation['peer']['peer'], None)



from django.contrib.auth.models import AnonymousUser, Permission, User
from django.http import HttpResponse
from django.test import TestCase
from django.views import generic

from ai_django_core.tests.mixins import RequestProviderMixin
from ai_django_core.view_layer.mixins import DjangoPermissionRequiredMixin


class MetaDjangoPermissionRequiredMixinTest(RequestProviderMixin, TestCase):
    class TestViewNoPerms(DjangoPermissionRequiredMixin, generic.View):
        pass

    class TestViewSinglePerm(DjangoPermissionRequiredMixin, generic.View):
        permission_list = ['auth.change_user']

        def get(self, *args, **kwargs):
            return HttpResponse(status=200)

    class TestViewMultiplePerms(DjangoPermissionRequiredMixin, generic.View):
        permission_list = ['auth.change_user', 'auth.add_user']

        def get_login_url(self):
            return 'login/'

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.permission = Permission.objects.get_by_natural_key(app_label='auth', codename='change_user', model='user')

    def setUp(self) -> None:
        super().setUp()
        self.user = User.objects.create(username='test_user', email='test.user@ai-django-core.com')

    def test_get_login_url(self):
        self.assertEqual(self.TestViewMultiplePerms().get_login_url(), 'login/')

    def test_permissions_are_set_validation(self):
        with self.assertRaises(RuntimeError):
            self.TestViewNoPerms()

    def test_has_permissions_correct_permission(self):
        self.user.user_permissions.add(self.permission)

        self.assertTrue(self.TestViewSinglePerm().has_permissions(self.user))

    def test_has_permissions_missing_permission(self):
        self.assertFalse(self.TestViewSinglePerm().has_permissions(self.user))

    def test_has_permissions_multiple_permissions_one_missing(self):
        self.user.user_permissions.add(self.permission)

        self.assertFalse(self.TestViewMultiplePerms().has_permissions(self.user))

    def test_passes_login_barrier_no_login_required(self):
        view = self.TestViewSinglePerm()
        view.login_required = False

        self.assertTrue(view.passes_login_barrier(self.user))

    def test_passes_login_barrier_user_logged_in(self):
        self.assertTrue(self.TestViewSinglePerm().passes_login_barrier(self.user))

    def test_passes_login_barrier_user_not_logged_in(self):
        self.assertFalse(self.TestViewSinglePerm().passes_login_barrier(AnonymousUser()))

    def test_has_permissions_is_superuser_is_not_blocked(self):
        self.user.is_superuser = True
        self.assertTrue(self.TestViewSinglePerm().has_permissions(self.user))

    def test_has_permissions_django_superuser_is_always_allowed(self):
        self.user.is_superuser = True
        self.assertTrue(self.TestViewSinglePerm().has_permissions(self.user))

    def test_dispatch_lockout_on_missing_permissions(self):
        response = self.TestViewSinglePerm().dispatch(request=self.get_request(self.user))

        self.assertEqual(response.status_code, 403)

    def test_dispatch_working_on_having_permissions(self):
        self.user.user_permissions.add(self.permission)
        view = self.TestViewSinglePerm()
        response = view.dispatch(request=self.get_request(self.user))

        self.assertEqual(response.status_code, 200)



from django.utils.decorators import method_decorator
from graphene_django.forms.mutation import DjangoModelFormMutation
from graphql import GraphQLError
from graphql_jwt.decorators import login_required
from promise import Promise, is_thenable


class DjangoValidatedModelFormMutation(DjangoModelFormMutation):
    """
    Takes django ModelForm validations and passes them to GraphQL like any other validation error
    """

    class Meta:
        abstract = True

    @classmethod
    def mutate(cls, root, info, input):
        """
        Handle mutation logic.
        Most code derived one-to-one from base class.
        """

        def on_resolve(payload):
            try:
                payload.client_mutation_id = input.get("client_mutation_id")
            except Exception:
                raise Exception("Cannot set client_mutation_id in the payload object {}".format(repr(payload)))
            return payload

        result = cls.mutate_and_get_payload(root, info, **input)

        if result.errors:
            err_msg = ''
            for err in result.errors:
                err_msg += f"Field '{err.field}': {err.messages[0]} "

            raise GraphQLError(err_msg.strip())

        if is_thenable(result):
            return Promise.resolve(result).then(on_resolve)

        return on_resolve(result)


@method_decorator(login_required, name='perform_mutate')
class LoginRequiredDjangoModelFormMutation(DjangoValidatedModelFormMutation):
    """
    Ensures that you need to be logged in with GraphQL JWT (json web token) authentication
    """

    class Meta:
        abstract = True
